/* jshint esversion: 6, node: true, unused: true */
'use strict';

const moduleName = 'tray';

const Tray    = require('tray');
const Menu    = require('menu');
const fs      = require('fs');
const path    = require('path');
const app     = require('app');
const shell   = require('shell');
const dialog  = require('dialog');
const winston = require('winston');

const appPath = app.getAppPath();

const Queriable = require(path.join(appPath, 'queriable'));
const native    = require(path.join(appPath, 'native'));
const helper    = require(path.join(appPath, 'helper'));

const config    = require('./config');

const STATES = {
  disconnected: 'Disconnected',
  idle:         'Idle',
  scanning:     'Scanning',
  cleaning:     'Cleaning',
  syncing:      'Syncing'
};

const VersionHelper = require('./version_helper');

function ask(question, title, options) {
  var idxCancel = options.indexOf('Cancel');
  var idxNo     = options.indexOf('No');

  function executor(resolve, reject) {
    function callback(btnIndex) {
      if (btnIndex !== idxCancel && btnIndex !== idxNo) {
        resolve();
      } else {
        reject();
      }
    }

    dialog.showMessageBox({
      type:     'question',
      message:  question,
      title:    title,
      buttons:  options
    }, callback);
  }

  return new Promise(executor);
}

function Move(files, syncFolderPath) {

  function move(transition) {
    winston.info(`[${moduleName}] RENAME`, transition);

    return new Promise(function(resolve, reject) {
      fs.rename(transition.from, transition.to, function(err) {
        if (err) {
          reject(err);
        } else {
          resolve();
        }
      });
    });
  }

  function prepare(sourcePath) {
    var source = {
      path:  sourcePath,
      isDirectory: undefined
    };
    var basename    = path.basename(sourcePath);
    var destination = {
      path:  path.join(syncFolderPath, basename),
      isDirectory: undefined
    };

    return helper.stat(source.path).
    then(function(stats) {
      source.isDirectory = stats.isDirectory();

      function executor(resolve) {
        helper.stat(destination.path).
        then(function(stats) {
          destination.isDirectory = stats.isDirectory();
          resolve();
        }).
        catch(resolve);
      }

      return new Promise(executor);
    }).
    then(function() {
      var resolvedPromise = Promise.resolve({
        from: source.path,
        to: destination.path
      });

      var differInType = source.isDirectory !== destination.isDirectory;
      if (differInType) return resolvedPromise;

      var type     = source.isDirectory ? 'Folder' : 'File';
      var question = `Your Protosync folder already contains a ${type.toLowerCase()} named "${basename}". Do you want to replace it?`;
      var title    = `${type} already exists`;
      var options  = [
        `Replace ${type.toLowerCase()}`,
        'Cancel'
      ];

      return ask(question, title, options).then(function() {
        return resolvedPromise;
      });
    });
  }

  function run() {
    if (files.length === 0) return;

    var file = files.shift();
    prepare(file).then(move).then(run).catch(function(error) {
      if (error) {
        winston.error(`[${moduleName}] ${error}`);
        dialog.showErrorBox('Oops', error);
        files.length = 0;
        // TODO Question user what to do:
        // options = ['Retry', 'Skip', 'Abort']
        // 'Retry' => files.unshift(file); run(); break;
        // 'Skip'  => run(); break;
        // 'Abort' => files.length = 0; break;
      }
    });
  }

  this.run = run;
}

function ProtoTray(options) {

  var protoTray = this;

  var menuItems = buildItems(options);
  var menuItemsList = new Queriable([
    menuItems.login,
    menuItems.statusIdle,
    menuItems.separator, // --------
    menuItems.openSOUL,
    // menuItems.openRemotely,
    menuItems.openLocally,
    menuItems.separator, // --------
    menuItems.toggleLaunchAtLogin,
    menuItems.resetConfiguration,
    menuItems.separator, // --------
    menuItems.version,
    menuItems.quit
  ]);

  var versionHelper = new VersionHelper();

  function redrawTray() {
    var contextMenu = Menu.buildFromTemplate(menuItemsList.collection);
    contextMenu.items.forEach(function(item) {
      if (item.type === 'separator') return;
      menuItemsList.findBy({ label: item.label }).target = item;
    });
    trayIcon.setContextMenu(contextMenu);
  }

  versionHelper.on('updateAvailable', function(version) {
    menuItems.version.label = `Update to ProtoSync ${ version }`;
    menuItems.version.enabled = true;

    redrawTray();
  });

  var trayIcon;

  function initialize(values) {
    menuItems.toggleLaunchAtLogin.checked = values[0];
    menuItems.version.label = `ProtoSync ${ values[1] }`;

    trayIcon = drawTray(menuItemsList);

    // TODO we should work with promises here...
    // set state again in case it was set before the tray was initialized
    if (state !== STATES.disconnected) {
      state = STATES.disconnected;
      connect();
    }

    // TODO replace
    function handleDrop(files) {
      if (protoTray.authenticated) {
        new Move(files, options.syncFolderPath).run();
      } else {
        dialog.showErrorBox('Login required', 'Please login first.');
      }
    }
    trayIcon.on('drop-files', function(e, files) {
      setTimeout(function() { handleDrop(files); }, 0);
    });
    // with
    native.registerDropHandle(trayIcon);
  }

  // waits for state of launchd of agent/service so we can initialize the
  // checkbox correctly
  Promise.
    all([ native.isLaunchedAtLogin, versionHelper.getCurrentVersion ]).
    then(initialize);

  function show(item) {
    item.visible = true;
    if (item.target) item.target.visible = true;
  }
  function hide(item) {
    item.visible = false;
    if (item.target) item.target.visible = false;
  }
  function enable(item) {
    item.enabled = true;
    if (item.target) item.target.enabled = true;
  }
  protoTray.enable = function(itemName) {
    enable(menuItems[itemName]);
  };
  function disable(item) {
    item.enabled = false;
    if (item.target) item.target.enabled = false;
  }
  protoTray.disable = function(itemName) {
    disable(menuItems[itemName]);
  };

  function disconnect() {
    protoTray.authenticated = false;

    disable(menuItems.openSOUL);
    disable(menuItems.openLocally);
    disable(menuItems.openRemotely);
    show(menuItems.login);
    hide(menuItems.resetConfiguration);

    setState(STATES.disconnected);
  }
  function connect() {
    protoTray.authenticated = true;

    enable(menuItems.openSOUL);
    enable(menuItems.openLocally);
    enable(menuItems.openRemotely);
    hide(menuItems.login);
    show(menuItems.resetConfiguration);

    setState(STATES.idle);
  }

  // expose public interface
  protoTray.appIcon        = trayIcon;
  protoTray.menuItems      = menuItems;
  protoTray.authenticated  = false;
  protoTray.loggedInState  = connect;
  protoTray.loggedOutState = disconnect;

  // TODO move this to main when we can do NetworkManager.getConnection().isConnected()
  const quitButtons = {
    CANCEL: {
      id: 0,
      label: "Cancel"
    },
    ALWAYS_QUIT: {
      id: 1,
      label: "Quit And Don't Ask Again"
    },
    QUIT: {
      id: 2,
      label: "Quit"
    }
  };
  // ask before quitting when connected
  app.on('before-quit', function(e) {
    winston.debug(`[${ moduleName }] Quitting...`, { state: state });
    winston.debug(config.json);

    if (state === STATES.disconnected) return app.exit(0);
    if (config.forceQuit) return app.exit(0);

    e.preventDefault();

    dialog.showMessageBox({
      type:    "warning",
      title:   `Close ${ app.getName() }?`,
      message: `Your files will not be kept in sync when ${ app.getName() } is closed.`,
      buttons: Object.keys(quitButtons).map(function(btnKey) {
        return quitButtons[btnKey].label;
      })
    }, function(response) {
        if (response === quitButtons.CANCEL.id) return;
        if (response === quitButtons.ALWAYS_QUIT.id) config.forceQuit = true;

        app.exit(0);
      }
    );
  });

  var state = STATES.disconnected;
  var stateItemsList = menuItemsList.withProperty('state');
  function setState(newState) {
    if (state === newState) return;

    var stateItem = stateItemsList.findBy({ state: newState });
    if (stateItem) {
      state = newState;

      stateItemsList.withProperty('target').mapTo('target').set({ visible: false });
      show(stateItem);

      // well, electron on linux...
      if (process.platform === 'linux') trayIcon.setContextMenu(trayIcon.menu);

      winston.debug(`[${moduleName}]`, { state: state, activity: activity });
    }
  }
  protoTray.setState = function(newState) { setState(STATES[newState]); };

  var activity = 0;
  protoTray.setActivity = function setActivity(newActivity) {
    if (activity === newActivity) return;
    activity = newActivity;

    var iconPath = native.getTrayImagePath(newActivity);
    trayIcon.setImage(iconPath);


    if (activity) { protoTray.setState(STATES.syncing); }   
  };

  protoTray.setStatusText = function setStatusText(text) {
    menuItems.statusIdle.label = text;
    redrawTray();
  }
}

function drawTray(itemsList) {
  var tray = new Tray(native.getTrayImagePath(0));
  tray.setToolTip(app.getName());

  var contextMenu = Menu.buildFromTemplate(itemsList.collection);
  contextMenu.items.forEach(function(item) {
    if (item.type === 'separator') return;
    itemsList.findBy({ label: item.label }).target = item;
  });
  tray.setContextMenu(contextMenu);

  return tray;
}

function buildItems(options) {
  return {
    login: {
      label: 'Setup Required',
      click: options.login,
      state: STATES.disconnected
    },
    statusDisconnected: {
      label: 'Disconnected',
      enabled: false,
      visible: false,
    },
    statusIdle: {
      // label: 'Idle',
      label: 'Ready',
      enabled: false,
      visible: false,
      state: STATES.idle
    },
    statusScanning: {
      label: 'Scanning',
      enabled: false,
      visible: false,
      state: STATES.scanning
    },
    statusCleaning: {
      label: 'Cleaning',
      enabled: false,
      visible: false,
      state: STATES.cleaning
    },
    statusSyncing: {
      label: 'Syncing',
      enabled: false,
      visible: false,
      state: STATES.syncing
    },
    openLocally: {
      label: 'Open Folder',
      click: options.openSyncFolder,
      enabled: false
    },
    openSOUL: {
      label: 'Open SOUL',
      enabled: false,
      click: function() {
        var url = `https://${config.boxName}/`;
        shell.openExternal(url);
      }
    },
    openRemotely: {
      label: 'Open Folder in Web',
      enabled: false,
      click: function() {
        var url = `https://${config.boxName}/#/my_files/files?owner_type=User&owner_id=${config.uid}&filter=protosync&path=%2Fprotosync%2Ffiles`;
        shell.openExternal(url);
      }
    },
    toggleLaunchAtLogin: {
      type: 'checkbox',
      label: 'Launch At Login',
      click: native.toggleLaunchAtLogin,
      checked: false
    },
    resetConfiguration: {
      label: 'Reset Configuration',
      click: function() {
        dialog.showMessageBox({
              type:    "warning",
              title:   "Really reset configuration?",
              message: "This will reset your configuration and requires you to re-enter your credentials.",
              buttons: ["Cancel", "Reset Configuration"]},
              function(response) {
                // response == 0 means, the user clicked cancel
                if (response !== 0) options.login();
              }
          );
      },
      visible: false
    },
    version: {
      label: 'Loading',
      enabled: false, // RADAR enable this when there is an update available?
      click: function() { shell.openExternal(native.downloadURL); }
    },
    quit: {
      label: 'Quit',
      accelerator: 'CommandOrControl+Q',
      click: function() { app.quit(); }
    },
    separator: {
      type: 'separator'
    }
  };
}

module.exports = ProtoTray;
