/* jshint esversion: 6, node: true, unused: true */

/* VersionHelper
 * 
 * const VersionHelper = require('./version_helper');
 * 
 * Emits the updateAvailable event if the current remote version is higher than
 * the current installed version.
 * 
 * var versionHelper = new VersionHelper({ updateURL: url });
 * versionHelper.on('updateAvailable', callback);
 */

const EventEmitter = require('events');
const util         = require('util');
const app          = require('electron').app;
const path         = require('path');

const ready  = require('./ready');
const helper = require('./helper');

const defaults = {
  updateInterval: 4*60*60*1000, // 4h
  updateURL: `http://packages2.protonet.info/protosync/current/VERSION.${ process.platform }`
};

// allows 1.0-beta1 to be less than 1.0
function decompose(version) {
  var parts = version.split('-');
  return { main: parts[0], sub: parts[1] };
}
function compareVersion(v1, v2) {
  v1 = decompose(v1);
  v2 = decompose(v2);

  if (v1.main < v2.main)    return -1;
  if (v1.main > v2.main)    return  1;
  if (v1.sub === v2.sub)    return  0;
  if (v2.sub === undefined) return -1;
  if (v1.sub < v2.sub)      return -1;

  return 1;
}

function trimRight(string) {
  string = string.trimRight();
  return Promise.resolve(string);
}

const versionPath = path.join(app.getAppPath(), 'VERSION');
var readCurrentVersion = helper.read(versionPath).then(trimRight);
ready.when(function(resolve, reject) {
  readCurrentVersion.then(resolve, reject);
});

function VersionHelper(options) {
  EventEmitter.call(this);

  if (options === undefined) options = 0;
  var updateInterval = options.updateInterval || defaults.updateInterval;
  var updateURL      = options.updateURL || defaults.updateURL;

  var instance = this;

  function compareLocalToRemoteVersion(resolve) {
    var prerequisites = [ readCurrentVersion, fetchLatestVersion() ];
    Promise.all(prerequisites).then(function(versions) {
      versions = versions.map(function(v) { return v; });

      if (compareVersion.apply(instance, versions) < 0) {
        resolve(versions[1]);
      } else {
        resolve(false);
      }
    });
  }

  function fetchLatestVersion() {
    return helper.getText(updateURL).then(trimRight);
  }
  function isUpdateAvailable() {
    return new Promise(compareLocalToRemoteVersion);
  }

  var nextCheck;
  function checkForUpdate() {
    isUpdateAvailable().then(function(version) {
      if (version) instance.emit('updateAvailable', version);
      nextCheck = setTimeout(checkForUpdate, updateInterval);
    });
  }
  app.on('quit', function() { clearTimeout(nextCheck); });

  ready.then(checkForUpdate);

  // public API
  this.setUpdateInterval = function(ms) { updateInterval = ms; };
  this.getCurrentVersion = readCurrentVersion;
  this.getLatestVersion  = fetchLatestVersion;
  this.isUpdateAvailable = isUpdateAvailable;
}
util.inherits(VersionHelper, EventEmitter);

module.exports = VersionHelper;
